
module Mpi where

infix 4  ><
infix 4  -|-

-- (1) Product -----------------------------------------------------------------

(><) :: (a -> b) -> (c -> d) -> (a,c) -> (b,d)
(f >< g)(a,b) = (f a, g b)

split :: (a -> b) -> (a -> c) -> a -> (b,c)
split f g x = (f x, g x)

-- the 0-adic split 

(!) :: a -> ()
(!) = const ()

-- Renamings:

p1      = fst
p2      = snd

-- (2) Coproduct ---------------------------------------------------------------

-- either is predefined

(-|-) :: (a -> b) -> (c -> d) -> Either a c -> Either b d
(f -|- g) (Left  a) = Left (f a)
(f -|- g) (Right b) = Right (g b)

-- Renamings:

i1      = Left
i2      = Right

-- (3) Exponentiation ---------------------------------------------------------

-- curry is predefined

ev :: (a -> b,a) -> b
ev = uncurry ($)

expn f = \g -> f . g

-- (4) Others -----------------------------------------------------------------

--const :: a -> b -> a st const a x = a is predefined

grd :: (a -> Bool) -> a -> Either a a
grd p x = if p x then Left x else Right x

-- (5) Natural isomorphisms ----------------------------------------------------

swap :: (a,b) -> (b,a)
swap (a,b) = (b,a)

assocr = split ( fst . fst ) (split ( snd . fst ) snd )

-- (6) Basic bifunctors --------------------------------------------------------

class BiFunctor f where
      bmap :: (a -> c) -> (b -> d) -> (f a b -> f c d)

instance BiFunctor Either where
    bmap f g = f -|- g

instance BiFunctor (,) where
    bmap f g (a,b) = (f >< g)(a,b)


